# 1 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F" 1
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright 2000-2025 CP2K developers group <https://cp2k.org>                                   !
!                                                                                                  !
!   SPDX-License-Identifier: GPL-2.0-or-later                                                      !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Input control types for NEGF/SMEAGOL transport calculations.
!> \author Sergey Chulkov
!> \author Christian Ahart
!> \author Clotilde Cucinotta
! **************************************************************************************************

MODULE smeagol_control_types
   USE cp_units, ONLY: cp_unit_from_cp2k
   USE input_constants, ONLY: smeagol_bulklead_leftright, &
                              smeagol_gridmethod_traditional, &
                              smeagol_integraltype_gauss_legendre, &
                              smeagol_runtype_bulktransport, &
                              smeagol_runtype_emtransport
   USE input_section_types, ONLY: section_vals_get, &
                                  section_vals_get_subs_vals, &
                                  section_vals_type, &
                                  section_vals_val_get
   USE kinds, ONLY: default_string_length, &
                    dp
   USE physcon, ONLY: kelvin
   USE string_utilities, ONLY: integer_to_string
   USE util, ONLY: sort
#include "./base/base_uses.f90"
# 1 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/input_cp2k_smeagol.fypp" 1
# 7 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/input_cp2k_smeagol.fypp"
# 325 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/input_cp2k_smeagol.fypp"
# 33 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F" 2

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'smeagol_control_types'

   PUBLIC :: smeagol_control_type
   PUBLIC :: smeagol_control_create, smeagol_control_release, read_smeagol_control

! **************************************************************************************************
!> \brief SMEAGOL-related input parameters
! **************************************************************************************************
   TYPE smeagol_control_type
      LOGICAL                                            :: smeagol_enabled = .FALSE.

      !> type of calculation
      INTEGER                                            :: run_type = smeagol_runtype_bulktransport

      !> regression test mode
      LOGICAL                                            :: do_regtest = .FALSE.

      !> current-induced forces. It is set automatically based on GLOBAL/run_type
      LOGICAL                                            :: emforces = .FALSE.

      !> scale factor to convert from CP2K (Hartree) to SMEAGOL (Rydberg) default energy unit
      REAL(kind=dp)                                      :: to_smeagol_energy_units = 2.0_dp

      !> number of cell images along i and j cell vectors
      INTEGER, DIMENSION(2)                              :: n_cell_images = (/1, 1/)

      !> what lead (bulk transport calculation)
      INTEGER                                            :: lead_label = smeagol_bulklead_leftright

      !> The length of the SMEAGOL project name is limited by 20 characters (hardcoded in SMEAGOL)
      CHARACTER(len=20)                                  :: project_name = "PROJECT_NAME"

      TYPE(smeagol_aux_control_type), POINTER            :: aux => NULL()

   END TYPE smeagol_control_type

! **************************************************************************************************
!> \brief SMEAGOL-related auxiliary input parameters. They remain unallocated when
!>        SMEAGOL support is disabled.
! **************************************************************************************************
   TYPE smeagol_aux_control_type
      !> MD or GEO_OPT iteration. In contrast with other components of smeagol_control_type that are read from the input file,
      !> these variables are initialised at the first NEGF iteration.
      !>
      !> index of MD or GEO_OPT iteration level.
      !> 0 if there is neither MD nor GEO_OPT iteration level (e.g. single-point energy calculation).
      !> -1 if it is not initialised by run_smeagol_emtrans() subroutine
      INTEGER                                            :: md_iter_level = -1
      !> The starting step value for MD / GEO_OPT iterations. The default value 0 can be overrited via STEP_START_VAL input keyword.
      INTEGER                                            :: md_first_step = 0

      !> BS.SubSystemsDelta(1:BS.Subsystems)
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: deltabss_bs
      !> BS.SubSystemsBoundaries(1:BS.Subsystems, 1:2)
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: nebss_bs
      !>
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atomlist_bs

      REAL(kind=dp)                                      :: temperature = 300.0_dp/kelvin

      ! reademtr()
# 99 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: TrCoeff = .false.
# 99 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: skiplast = .false.
# 99 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: CalcIETS = .false.
# 101 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 103 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_maxR = .FALSE.
# 105 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 107 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: dsigmade = .false.
# 107 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: calcmpsh = .false.
# 109 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 111 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: COOPCalculate = .false.
# 113 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 115 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: NEnergR = 288
# 115 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: NEnergIC = 32
# 115 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: NEnergIL = 16
# 115 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: NPoles = 16
# 115 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: SpinCL = 0
# 115 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: AtmLVCte = 1
# 115 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: AtmRVCte = -1
# 115 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: NSlices = 1
# 115 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: nlinmix = 1
# 115 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: tmdskip = 0
# 115 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: tmdsampling = 1
# 117 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 119 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: NeneT = 100
# 119 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: rmoinit = 1
# 119 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: rmolast = 1
# 119 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: rorbinit = 1
# 119 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: rorblast = 1
# 121 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 123 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: COOPNumberOfBonds = 0
# 125 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 127 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: Delta = 1.0e-10_dp
# 129 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 131 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: maxkappa = 0.0_dp
# 133 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 135 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: VBias = 0.0_dp*0.5_dp
# 135 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: EnergLB = -7.0_dp*0.5_dp
# 135 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: minL = 0.0_dp
# 137 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 139 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: maxR = 0.0_dp
# 141 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 143 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: TEnergI = -0.3675_dp*0.5_dp
# 143 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: TEnergF = -0.3675_dp*0.5_dp
# 143 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: vgate = 0.0_dp*0.5_dp
# 145 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 147 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: zgater = 0.0_dp
# 147 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: zgatel = 0.0_dp
# 147 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: alphag = 1.0_dp
# 149 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! ReadOptionsNEGF_DFT()
# 152 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: timereversal = .false.
# 152 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: TRCAddVCDFT = .false.
# 152 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: AddVgCompensatingIsolatedLocalCharges = .true.
# 152 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: Add_Rho_Gate = .false.
# 152 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: debug_gate = .false.
# 152 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: empdoskso = .false.
# 152 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: emtimings = .false.
# 154 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 156 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_RGlxmin = .FALSE.
# 156 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_RGlxmax = .FALSE.
# 156 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_RGlymin = .FALSE.
# 156 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_RGlymax = .FALSE.
# 156 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_RGlzmin = .FALSE.
# 156 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_RGlzmax = .FALSE.
# 158 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 160 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: ik_select = 0
# 162 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 164 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: Q_gate = 0.0_dp
# 166 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 168 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: RGlxmin = 0.0_dp
# 168 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: RGlxmax = 0.0_dp
# 168 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: RGlymin = 0.0_dp
# 168 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: RGlymax = 0.0_dp
# 168 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: RGlzmin = 0.0_dp
# 168 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: RGlzmax = 0.0_dp
# 170 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 172 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: ZeemanBx = 0.0_dp*0.5_dp
# 172 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: ZeemanBy = 0.0_dp*0.5_dp
# 172 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: ZeemanBz = 0.0_dp*0.5_dp
# 174 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! emtrans_options()
      INTEGER                                            :: gridmethod = smeagol_gridmethod_traditional
      INTEGER                                            :: integraltype = smeagol_integraltype_gauss_legendre

# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: pdosgs = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: skipright = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: sigmatodisk = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: bs_add = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: bs_writetrc = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: writetk = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: setene = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: overwritehs = .true.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: transmembed = .true.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: writeevsigma = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: skiptransm = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: deauto = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: geterrsigma = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: getsigma = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: trcef = .true.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: RhoSetZeroIfHZero = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: m_usehinv = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: m_usevinv = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: m_callsvd = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: m_complexbands = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: m_dosleads = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: leadsdos = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: leadspdos = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: emldos = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: emldos2 = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: curr_distKEne = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: curr_distK = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: curr_dist = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: curr_distTRC = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: em_NonSCF = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: TransmissionRL = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: emSTTk = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: GetT0S = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: TransmissionChannels = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: TransmissionChannelsWFS = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: TransmissionMatrix = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: TransmissionMatrixWFS = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: TransmissionMatrixPDOS = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: EM_EndCode1 = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: TransmOverk = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: emdos = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: empdos = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: empdosk = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: WriteIkTrcK = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: cbspart = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: evec = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: negfon = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: outinfo = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: WriteNk = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: Set_RhoBoundary_Leads = .true.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: Set_RhoBoundaryOverlap_Leads = .true.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: Set_HBoundary_Leads = .true.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: Set_HLR_Zero = .true.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: SkipNonEquilibriumRho = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: SkipEquilibriumRho = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: ComputeImpurityGfMatsubara = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: PrintImpurityGfMatsubara = .false.
# 180 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: CallImpuritySolver = .false.
# 182 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 184 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_nprocs_hs = .FALSE.
# 184 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_Sigma_NxLeft = .FALSE.
# 184 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_Sigma_NyLeft = .FALSE.
# 184 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_Sigma_NxRight = .FALSE.
# 184 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_Sigma_NyRight = .FALSE.
# 186 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 188 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_GetRhoSingleLead = .FALSE.
# 190 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 192 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_deltamin = .FALSE.
# 194 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 196 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: nprocs_inverse = 1
# 196 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: Sigma_Nx = 1
# 196 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: Sigma_Ny = 1
# 198 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 200 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: nprocs_hs = 1
# 200 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: Sigma_NxLeft = 1
# 200 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: Sigma_NyLeft = 1
# 200 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: Sigma_NxRight = 1
# 200 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: Sigma_NyRight = 1
# 202 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: maxdepth = 15
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: ndivisions = 9
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: storesigma = 0
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: bs_method = 0
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: bs_nmid = 0
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: NRunSigmaMax = 3
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: bsrun = 0
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: bsskip = 1
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: bseskip = 1
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: bssc = 1
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: inversion_solver = 1
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: onekp = 0
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: nbss = 2
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: nSplit = 0
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: TransmissionChannelsWFSSkipEne = 100
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: TransmissionChannelsWFSSkipKP = 100
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: TransmissionMatrixPDOSNWrite = 20
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: TransmissionMatrixiSetPhase = -1
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: MinChannelIndex = 1
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: MaxChannelIndex = 0
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: N_Replace_L = 1
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: N_Replace_R = 1
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: curr_nb1 = -1
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: curr_nb2 = -1
# 204 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: NParallelK = 1
# 206 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 208 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: GetRhoSingleLead = 0
# 210 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 212 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         INTEGER                                            :: sigma_method = 1
# 214 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: bs_tol = 1.0e-05_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: bs_min  = 0.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: HSTol  = 1.0e-10_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: deltaimag  = 0.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: deltaini  = 1.0e-03_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: deltatode  = 1.5_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: critam  = 2.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: setemin  = -1.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: setemax  = 1.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: tolab  = 1.0e-06_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: weightrho  = 0.5_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: ThetaLeadsL  = 0.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: PhiLeadsL  = 0.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: ThetaLeadsR  = 0.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: PhiLeadsR  = 0.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: deltaimagtrc  = 0.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: m_tolki  = 1.0e-06_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: m_svdtolmax  = 5.0e-12_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: m_svdtolmin  = 1.0e-15_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: m_dsigmamax  = 1.0e-05_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: m_rnoise  = 1.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: m_skipsvd  = 1.0e+03_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: m_svdtolzi  = 1.0e-08_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: SigmaWideBand  = 0.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: emSTTkScale  = 1.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: nlSetZeroRatio  = 1.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: nrSetZeroRatio  = 1.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: curr_fl_L  = 1.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: curr_fr_L  = 0.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: curr_fl_R  = 0.0_dp
# 216 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: curr_fr_R  = 1.0_dp
# 218 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 220 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: deltamin = 0.0_dp
# 222 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 224 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: tolsigma = 1.0e-05_dp*0.5_dp
# 224 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: trcde = 0.0_dp*0.5_dp
# 226 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! interface options
# 229 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: HartreeLeadsLeft = 0.0_dp
# 229 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: HartreeLeadsRight = 0.0_dp
# 231 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
# 232 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         REAL(kind=dp)                                      :: HartreeLeadsBottom = 0.0_dp*0.5_dp
# 234 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
# 235 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         LOGICAL                                            :: isexplicit_HartreeLeadsBottom = .FALSE.
# 237 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

   END TYPE smeagol_aux_control_type
CONTAINS

! **************************************************************************************************
!> \brief allocate control options for SMEAGOL calculation
!> \param smeagol_control an object to create
! **************************************************************************************************
   SUBROUTINE smeagol_control_create(smeagol_control)
      TYPE(smeagol_control_type), POINTER                :: smeagol_control

      CHARACTER(len=*), PARAMETER :: routineN = 'smeagol_control_create'

      INTEGER                                            :: handle

      CPASSERT(.NOT. ASSOCIATED(smeagol_control))
      CALL timeset(routineN, handle)

      ALLOCATE (smeagol_control)
#if defined(__SMEAGOL)
      ALLOCATE (smeagol_control%aux)
#else
      NULLIFY (smeagol_control%aux)
#endif

      CALL timestop(handle)
   END SUBROUTINE smeagol_control_create

! **************************************************************************************************
!> \brief release SMEAGOL control object
!> \param smeagol_control an object to release
! **************************************************************************************************
   SUBROUTINE smeagol_control_release(smeagol_control)
      TYPE(smeagol_control_type), POINTER                :: smeagol_control

      CHARACTER(len=*), PARAMETER :: routineN = 'smeagol_control_release'

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(smeagol_control)) THEN
         IF (ASSOCIATED(smeagol_control%aux)) THEN
            IF (ALLOCATED(smeagol_control%aux%nebss_bs)) DEALLOCATE (smeagol_control%aux%nebss_bs)
            IF (ALLOCATED(smeagol_control%aux%deltabss_bs)) DEALLOCATE (smeagol_control%aux%deltabss_bs)
            IF (ALLOCATED(smeagol_control%aux%atomlist_bs)) DEALLOCATE (smeagol_control%aux%atomlist_bs)

            DEALLOCATE (smeagol_control%aux)
         END IF
         DEALLOCATE (smeagol_control)
      END IF

      CALL timestop(handle)
   END SUBROUTINE smeagol_control_release

! **************************************************************************************************
!> \brief Read SMEAGOL-related input parameters.
!> \param smeagol_control SMEAGOL control parameters
!> \param smeagol_section SMEAGOL input section
! **************************************************************************************************
   SUBROUTINE read_smeagol_control(smeagol_control, smeagol_section)
      TYPE(smeagol_control_type), POINTER                :: smeagol_control
      TYPE(section_vals_type), POINTER                   :: smeagol_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_smeagol_control'

      CHARACTER(len=default_string_length)               :: project_name
      INTEGER                                            :: handle
      INTEGER, DIMENSION(:), POINTER                     :: n_cell_images_ptr

      CALL timeset(routineN, handle)

      CALL section_vals_val_get(smeagol_section, "_SECTION_PARAMETERS_", l_val=smeagol_control%smeagol_enabled)

      ! SMEAGOL project name
      CALL section_vals_val_get(smeagol_section, "PROJECT_NAME", c_val=project_name)
      smeagol_control%project_name = project_name
      IF (smeagol_control%smeagol_enabled .AND. smeagol_control%project_name /= project_name) THEN
         CALL cp_warn(__LOCATION__, &
                      "SMEAGOL limits the length of the project name by 20 characters. "// &
                      "The project name is truncated to '"//TRIM(smeagol_control%project_name)//"'.")
      END IF

      ! reademtr() enum keywords
      CALL section_vals_val_get(smeagol_section, "RUN_TYPE", i_val=smeagol_control%run_type)

#if !defined(__SMEAGOL)
      IF (smeagol_control%run_type == smeagol_runtype_emtransport) THEN
         CALL cp_abort(__LOCATION__, &
                       "CP2K was compiled with no SMEAGOL support. SMEAGOL RUN_TYPE EMTransport is not available")
      END IF
#endif

      CALL section_vals_val_get(smeagol_section, "REGRESSION_TEST", l_val=smeagol_control%do_regtest)

      CALL section_vals_val_get(smeagol_section, "BulkLead", i_val=smeagol_control%lead_label)

      ! NOTE: keyword NIMAGES_IJ seems to be identical to ignored SMEAGOL keywords BulkTransvCellSizeX and BulkTransvCellSizeY
      NULLIFY (n_cell_images_ptr)
      CALL section_vals_val_get(smeagol_section, "NIMAGES_IJ", i_vals=n_cell_images_ptr)
      smeagol_control%n_cell_images(1:2) = n_cell_images_ptr(1:2)

      ! Hartree > Rydberg scaling factor
      smeagol_control%to_smeagol_energy_units = cp_unit_from_cp2k(1.0_dp, "RY")

      IF (ASSOCIATED(smeagol_control%aux)) CALL read_smeagol_aux_control(smeagol_control%aux, smeagol_section)

      CALL timestop(handle)
   END SUBROUTINE read_smeagol_control

! **************************************************************************************************
!> \brief Read SMEAGOL-related auxiliary input parameters.
!> \param smeagol_control SMEAGOL control parameters
!> \param smeagol_section SMEAGOL input section
! **************************************************************************************************
   SUBROUTINE read_smeagol_aux_control(smeagol_control, smeagol_section)
      TYPE(smeagol_aux_control_type), POINTER            :: smeagol_control
      TYPE(section_vals_type), POINTER                   :: smeagol_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_smeagol_aux_control'

      CHARACTER(len=default_string_length)               :: nvals_str
      INTEGER                                            :: handle, i, n, nrep, n_unique
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: ivec, indices
      INTEGER, DIMENSION(:), POINTER                     :: iptr
      LOGICAL                                            :: is_explicit, do_abort
      REAL(kind=dp), DIMENSION(:), POINTER               :: rptr
      TYPE(section_vals_type), POINTER                   :: subsection

      CALL timeset(routineN, handle)

      smeagol_control%md_iter_level = -1
      smeagol_control%md_first_step = 0

      CALL section_vals_val_get(smeagol_section, "TEMPERATURE", r_val=smeagol_control%temperature)

      ! reademtr() logical keywords
      ! The following keywords (variables) that are part of the reademtr_local_llist list are read but unused
      ! CB.WriteComplexBands (WriteComplexBands = .FALSE.) unimplemented.
      ! Presumably, some of these keywords should be removed.
# 378 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "TrCoefficients", l_val=smeagol_control%TrCoeff)
# 378 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "SkipLastIter", l_val=smeagol_control%skiplast)
# 378 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "CalcIETS", l_val=smeagol_control%CalcIETS)
# 380 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
# 381 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.DSigmaDE", l_val=smeagol_control%dsigmade)
# 381 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "CalcMPSH", l_val=smeagol_control%calcmpsh)
# 383 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
# 384 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.COOPCalculate", l_val=smeagol_control%COOPCalculate)
# 386 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! reademtr() integer keywords
# 389 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "NEnergReal", i_val=smeagol_control%NEnergR)
# 389 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "NenergImCircle", i_val=smeagol_control%NEnergIC)
# 389 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "NenergImLine", i_val=smeagol_control%NEnergIL)
# 389 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "NPoles", i_val=smeagol_control%NPoles)
# 389 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "SpinConfLeads", i_val=smeagol_control%SpinCL)
# 389 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "AtomLeftVcte", i_val=smeagol_control%AtmLVCte)
# 389 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "AtomRightVcte", i_val=smeagol_control%AtmRVCte)
# 389 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "NSlices", i_val=smeagol_control%NSlices)
# 389 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "NumberLinearMix", i_val=smeagol_control%nlinmix)
# 389 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "MD.TRCSkip", i_val=smeagol_control%tmdskip)
# 389 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "MD.TRCSampling", i_val=smeagol_control%tmdsampling)
# 391 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
# 392 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "NTransmPoints", i_val=smeagol_control%NeneT)
# 392 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "MPSHAtomFirst", i_val=smeagol_control%rmoinit)
# 392 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "MPSHAtomLast", i_val=smeagol_control%rmolast)
# 392 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "MPSHOrbFirst", i_val=smeagol_control%rorbinit)
# 392 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "MPSHOrbLast", i_val=smeagol_control%rorblast)
# 394 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
# 395 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.COOPNumberOfBonds", i_val=smeagol_control%COOPNumberOfBonds)
# 397 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! reademtr() real-valued keywords
# 400 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Delta", r_val=smeagol_control%Delta)
# 402 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
# 403 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "CB.MaxKappa", r_val=smeagol_control%maxkappa)
# 405 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! reademtr() physical-valued keywords
# 408 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "VBias", r_val=smeagol_control%VBias)
# 408 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EnergLowestBound", r_val=smeagol_control%EnergLB)
# 408 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "ZLeftVcte", r_val=smeagol_control%minL)
# 410 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
# 411 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "ZRightVcte", &
                                   r_val=smeagol_control%maxR, &
                                   explicit=smeagol_control%isexplicit_maxR)
# 415 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
# 416 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "InitTransmRange", r_val=smeagol_control%TEnergI)
# 416 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "FinalTransmRange", r_val=smeagol_control%TEnergF)
# 416 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "VGate", r_val=smeagol_control%vgate)
# 418 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
# 419 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "ZVGateR", r_val=smeagol_control%zgater)
# 419 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "ZVGateL", r_val=smeagol_control%zgatel)
# 419 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "ZBroadeningG", r_val=smeagol_control%alphag)
# 421 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! ReadOptionsNEGF_DFT() logical keywords
# 424 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TimeReversal", l_val=smeagol_control%timereversal)
# 424 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TRCAddVCDFT", l_val=smeagol_control%TRCAddVCDFT)
# 424 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.AddVgIsolatedLocalCharges", l_val=smeagol_control%AddVgCompensatingIsolated&
# 424 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
             &LocalCharges)
# 424 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.AddRhoGate", l_val=smeagol_control%Add_Rho_Gate)
# 424 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.DebugRhoGate", l_val=smeagol_control%debug_gate)
# 424 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EMPDOSKSO", l_val=smeagol_control%empdoskso)
# 424 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.Timings", l_val=smeagol_control%emtimings)
# 426 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! ReadOptionsNEGF_DFT() integer keywords
# 429 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Ik_Select", i_val=smeagol_control%ik_select)
# 431 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! ReadOptionsNEGF_DFT() real-valued keywords
# 434 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.NetRhoGateCharge", r_val=smeagol_control%Q_gate)
# 436 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! ReadOptionsNEGF_DFT() physical-valued keywords
# 439 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.RhoGateLxMin", &
                                   r_val=smeagol_control%RGlxmin, &
                                   explicit=smeagol_control%isexplicit_RGlxmin)
# 439 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.RhoGateLxMax", &
                                   r_val=smeagol_control%RGlxmax, &
                                   explicit=smeagol_control%isexplicit_RGlxmax)
# 439 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.RhoGateLyMin", &
                                   r_val=smeagol_control%RGlymin, &
                                   explicit=smeagol_control%isexplicit_RGlymin)
# 439 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.RhoGateLyMax", &
                                   r_val=smeagol_control%RGlymax, &
                                   explicit=smeagol_control%isexplicit_RGlymax)
# 439 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.RhoGateLzMin", &
                                   r_val=smeagol_control%RGlzmin, &
                                   explicit=smeagol_control%isexplicit_RGlzmin)
# 439 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.RhoGateLzMax", &
                                   r_val=smeagol_control%RGlzmax, &
                                   explicit=smeagol_control%isexplicit_RGlzmax)
# 443 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 445 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "ZeemanTermBx", r_val=smeagol_control%ZeemanBx)
# 445 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "ZeemanTermBy", r_val=smeagol_control%ZeemanBy)
# 445 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "ZeemanTermBz", r_val=smeagol_control%ZeemanBz)
# 447 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! emtrans_options() enum keywords
      CALL section_vals_val_get(smeagol_section, "EnergyGridType", i_val=smeagol_control%gridmethod)
      CALL section_vals_val_get(smeagol_section, "TypeOfIntegral", i_val=smeagol_control%integraltype)

      ! emtrans_options() logical keywords
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.PDOS", l_val=smeagol_control%pdosgs)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.SkipRight", l_val=smeagol_control%skipright)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.WriteToDisk", l_val=smeagol_control%sigmatodisk)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "BS.Add", l_val=smeagol_control%bs_add)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "BS.WriteTransmission", l_val=smeagol_control%bs_writetrc)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Current.WriteTransmission", l_val=smeagol_control%writetk)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "SetEnergyRange", l_val=smeagol_control%setene)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.OverwriteHS", l_val=smeagol_control%overwritehs)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "TransmissionEmbed", l_val=smeagol_control%transmembed)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.WriteEV", l_val=smeagol_control%writeevsigma)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.SkipTransmission", l_val=smeagol_control%skiptransm)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "TRCDEAuto", l_val=smeagol_control%deauto)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.CheckAccuracy", l_val=smeagol_control%geterrsigma)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.Write", l_val=smeagol_control%getsigma)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "TRCScaleEf", l_val=smeagol_control%trcef)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.SetRhoZeroIfHZero", l_val=smeagol_control%RhoSetZeroIfHZero)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.InvertH1", l_val=smeagol_control%m_usehinv)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.InvertV", l_val=smeagol_control%m_usevinv)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.CSVD", l_val=smeagol_control%m_callsvd)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.ComplexBands", l_val=smeagol_control%m_complexbands)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.DOSVV", l_val=smeagol_control%m_dosleads)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "TRC.LeadsDOS", l_val=smeagol_control%leadsdos)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "TRC.LeadsPDOS", l_val=smeagol_control%leadspdos)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.LDOS3", l_val=smeagol_control%emldos)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.LDOS", l_val=smeagol_control%emldos2)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.CurrentDensityK_Ene", l_val=smeagol_control%curr_distKEne)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.CurrentDensityK", l_val=smeagol_control%curr_distK)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.CurrentDensity", l_val=smeagol_control%curr_dist)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.CurrentDensityTRC", l_val=smeagol_control%curr_distTRC)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.NonSelfConsistentRun", l_val=smeagol_control%em_NonSCF)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TransmissionRL", l_val=smeagol_control%TransmissionRL)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.SpinTorqueK", l_val=smeagol_control%emSTTk)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TRCSpinComponents", l_val=smeagol_control%GetT0S)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TRCChannels", l_val=smeagol_control%TransmissionChannels)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TRCChannelsWFS", l_val=smeagol_control%TransmissionChannelsWFS)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TMatrix", l_val=smeagol_control%TransmissionMatrix)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TMatrixWFS", l_val=smeagol_control%TransmissionMatrixWFS)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TMatrixPDOS", l_val=smeagol_control%TransmissionMatrixPDOS)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.EndCode1", l_val=smeagol_control%EM_EndCode1)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "TransmissionOverK", l_val=smeagol_control%TransmOverk)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "TRC.EMDOS", l_val=smeagol_control%emdos)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "TRC.EMPDOS", l_val=smeagol_control%empdos)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "TRC.EMPDOSK", l_val=smeagol_control%empdosk)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "TRC.TrcOverKWriteIk", l_val=smeagol_control%WriteIkTrcK)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "CBS.SubSystem", l_val=smeagol_control%cbspart)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "CBS.PrintWaveFunction", l_val=smeagol_control%evec)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.OrderN", l_val=smeagol_control%negfon)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.OutputInfo", l_val=smeagol_control%outinfo)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.WriteNk", l_val=smeagol_control%WriteNk)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.SetRhoBoundaryFromLeads", l_val=smeagol_control%Set_RhoBoundary_Leads)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.SetRhoBoundaryOverlapFromLeads",&
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
             & l_val=smeagol_control%Set_RhoBoundaryOverlap_Leads)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.SetHBoundaryFromLeads", l_val=smeagol_control%Set_HBoundary_Leads)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.SetHLRZero", l_val=smeagol_control%Set_HLR_Zero)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.SkipNonEquilibriumRho", l_val=smeagol_control%SkipNonEquilibriumRho)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.SkipEquilibriumRho", l_val=smeagol_control%SkipEquilibriumRho)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.ComputeImpurityGfMatsubara", l_val=smeagol_control%ComputeImpurityGfMatsuba&
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
             &ra)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.PrintImpurityGfMatsubara", l_val=smeagol_control%PrintImpurityGfMatsubara)
# 454 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.CallImpuritySolver", l_val=smeagol_control%CallImpuritySolver)
# 456 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! emtrans_options() integer keywords
# 459 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.NProcessorsInverse", i_val=smeagol_control%nprocs_inverse)
# 459 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.Nx", i_val=smeagol_control%Sigma_Nx)
# 459 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.Ny", i_val=smeagol_control%Sigma_Ny)
# 461 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 463 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.NProcessorsDistributeMatrices", &
                                   i_val=smeagol_control%nprocs_hs, &
                                   explicit=smeagol_control%isexplicit_nprocs_hs)
# 463 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.NxLeft", &
                                   i_val=smeagol_control%Sigma_NxLeft, &
                                   explicit=smeagol_control%isexplicit_Sigma_NxLeft)
# 463 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.NyLeft", &
                                   i_val=smeagol_control%Sigma_NyLeft, &
                                   explicit=smeagol_control%isexplicit_Sigma_NyLeft)
# 463 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.NxRight", &
                                   i_val=smeagol_control%Sigma_NxRight, &
                                   explicit=smeagol_control%isexplicit_Sigma_NxRight)
# 463 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.NyRight", &
                                   i_val=smeagol_control%Sigma_NyRight, &
                                   explicit=smeagol_control%isexplicit_Sigma_NyRight)
# 467 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "AM.MaxDepth", i_val=smeagol_control%maxdepth)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "AM.NumberDivisions", i_val=smeagol_control%ndivisions)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "SIGMA.Save", i_val=smeagol_control%storesigma)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "BS.Method", i_val=smeagol_control%bs_method)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "BS.MiddleOrbital", i_val=smeagol_control%bs_nmid)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.NRerunSigmaMax", i_val=smeagol_control%NRunSigmaMax)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "BS.TypeOfRun", i_val=smeagol_control%bsrun)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "BS.Skip", i_val=smeagol_control%bsskip)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "BS.ESkip", i_val=smeagol_control%bseskip)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "BS.SetOccupation", i_val=smeagol_control%bssc)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.InverseSolver", i_val=smeagol_control%inversion_solver)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.OneKP", i_val=smeagol_control%onekp)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "BS.Subsystems", i_val=smeagol_control%nbss)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TRCSpinComponentsOribalSplit", i_val=smeagol_control%nSplit)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TRCChannelsWFS_SkipEne", i_val=smeagol_control%TransmissionChannelsWFSSkipE&
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
             &ne)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TRCChannelsWFS_SkipKP", i_val=smeagol_control%TransmissionChannelsWFSSkipKP)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TMatrixPDOSNWrite", i_val=smeagol_control%TransmissionMatrixPDOSNWrite)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TMatrixSetPhaseIndex", i_val=smeagol_control%TransmissionMatrixiSetPhase)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TRCMinChannelIndex", i_val=smeagol_control%MinChannelIndex)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.TRCMaxChannelIndex", i_val=smeagol_control%MaxChannelIndex)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.NSetHBoundaryFromLeadsL", i_val=smeagol_control%N_Replace_L)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.NSetHBoundaryFromLeadsR", i_val=smeagol_control%N_Replace_R)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.CurrentFluxOrbital1", i_val=smeagol_control%curr_nb1)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.CurrentFluxOrbital2", i_val=smeagol_control%curr_nb2)
# 469 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.ParallelOverKNum", i_val=smeagol_control%NParallelK)
# 471 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 473 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.LDOSLeadsProjection", &
                                   i_val=smeagol_control%GetRhoSingleLead, &
                                   explicit=smeagol_control%isexplicit_GetRhoSingleLead)
# 477 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 479 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.Method", i_val=smeagol_control%sigma_method)
# 481 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! emtrans_options() real-valued keywords
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "BS.Tolerance", r_val=smeagol_control%bs_tol)
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "BS.Minimum", r_val=smeagol_control%bs_min )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.HSTol", r_val=smeagol_control%HSTol )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.EImag", r_val=smeagol_control%deltaimag )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "AM.DeltaInit", r_val=smeagol_control%deltaini )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "AM.DeltaToDe", r_val=smeagol_control%deltatode )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "AM.Criterium", r_val=smeagol_control%critam )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "SetEmin", r_val=smeagol_control%setemin )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "SetEmax", r_val=smeagol_control%setemax )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.Tolab", r_val=smeagol_control%tolab )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.WeightRho", r_val=smeagol_control%weightrho )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.RotateSpinLeadsLeft.Theta", r_val=smeagol_control%ThetaLeadsL )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.RotateSpinLeadsLeft.Phi", r_val=smeagol_control%PhiLeadsL )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.RotateSpinLeadsRight.Theta", r_val=smeagol_control%ThetaLeadsR )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.RotateSpinLeadsRight.Phi", r_val=smeagol_control%PhiLeadsR )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "DeltaTransmission", r_val=smeagol_control%deltaimagtrc )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.Dkimag", r_val=smeagol_control%m_tolki )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.DSVDMax", r_val=smeagol_control%m_svdtolmax )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.DSVDMin", r_val=smeagol_control%m_svdtolmin )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.DMax", r_val=smeagol_control%m_dsigmamax )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.RNoise", r_val=smeagol_control%m_rnoise )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.SkipSVD", r_val=smeagol_control%m_skipsvd )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.SVDTolZero", r_val=smeagol_control%m_svdtolzi )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.WideBandImag", r_val=smeagol_control%SigmaWideBand )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.SpinTorqueKScale", r_val=smeagol_control%emSTTkScale )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.SetHLRZeroRatioLeft", r_val=smeagol_control%nlSetZeroRatio )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.SetHLRZeroRatioRight", r_val=smeagol_control%nrSetZeroRatio )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.CurrentFluxFL_L", r_val=smeagol_control%curr_fl_L )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.CurrentFluxFR_L", r_val=smeagol_control%curr_fr_L )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.CurrentFluxFL_R", r_val=smeagol_control%curr_fl_R )
# 484 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "EM.CurrentFluxFR_R", r_val=smeagol_control%curr_fr_R )
# 486 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

# 488 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "BS.Delta", &
                                   r_val=smeagol_control%deltamin, &
                                   explicit=smeagol_control%isexplicit_deltamin)
# 492 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! emtrans_options() physical-valued keywords
# 495 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "Sigma.WarnInOutput", r_val=smeagol_control%tolsigma)
# 495 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "TRCDE", r_val=smeagol_control%trcde)
# 497 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! interface physical-valued keywords
# 500 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "HartreeLeadsLeft", r_val=smeagol_control%HartreeLeadsLeft)
# 500 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "HartreeLeadsRight", r_val=smeagol_control%HartreeLeadsRight)
# 502 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
# 503 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"
         CALL section_vals_val_get(smeagol_section, "HartreeLeadsBottom", &
                                   r_val=smeagol_control%HartreeLeadsBottom, &
                                   explicit=smeagol_control%isexplicit_HartreeLeadsBottom)
# 507 "/builddir/build/BUILD/cp2k-2025.1-build/cp2k-2025.1/src/smeagol_control_types.F"

      ! Bound states
      IF (smeagol_control%nbss <= 0) THEN
         CALL cp_abort(__LOCATION__, &
                       "BS.Subsystems keyword should have a positive integer value.")
      END IF

      subsection => section_vals_get_subs_vals(smeagol_section, "BS.SubSystemsBoundaries")
      CALL section_vals_get(subsection, explicit=is_explicit)
      IF (is_explicit) THEN
         CALL section_vals_val_get(subsection, "_DEFAULT_KEYWORD_", n_rep_val=nrep, explicit=is_explicit)
         IF (.NOT. is_explicit) nrep = 0

         do_abort = .FALSE.
         IF (smeagol_control%nbss == nrep) THEN
            IF (is_explicit) THEN
               ALLOCATE (smeagol_control%nebss_bs(nrep, 2))
               DO i = 1, nrep
                  CALL section_vals_val_get(subsection, "_DEFAULT_KEYWORD_", i_rep_val=i, i_vals=iptr)
                  IF (SIZE(iptr) == 2) THEN
                     IF (MINVAL(iptr) > 0 .AND. iptr(1) <= iptr(2)) THEN
                        smeagol_control%nebss_bs(i, 1:2) = iptr(1:2)
                     ELSE
                        do_abort = .TRUE.
                        EXIT
                     END IF
                  ELSE
                     do_abort = .TRUE.
                     EXIT
                  END IF
               END DO
            END IF
         ELSE
            do_abort = .TRUE.
         END IF

         IF (do_abort) THEN
            CALL integer_to_string(smeagol_control%nbss, nvals_str)
            CALL cp_abort(__LOCATION__, &
                          "BS.SubSystemsBoundaries section is expected to have BS.Subsystems ("//TRIM(nvals_str)// &
                          ") lines with two positive integer numbers on each line.")
         END IF

      END IF

      CALL section_vals_val_get(smeagol_section, "BS.SubSystemsDelta", explicit=is_explicit)
      IF (is_explicit) THEN
         CALL section_vals_val_get(smeagol_section, "BS.SubSystemsDelta", r_vals=rptr)
         IF (smeagol_control%nbss /= SIZE(rptr)) THEN ! do negative deltas make sense ? .OR. MINVAL(rptr) < 0
            CALL integer_to_string(smeagol_control%nbss, nvals_str)
            CALL cp_abort(__LOCATION__, &
                          "BS.SubSystemsDelta keyword is expected to have BS.Subsystems ("//TRIM(nvals_str)//") real numbers.")
         END IF

         n = SIZE(rptr)
         ALLOCATE (smeagol_control%deltabss_bs(n))
         smeagol_control%deltabss_bs(1:n) = rptr(1:n)
         !ELSE allocate and set deltabss_bs() to deltamin
      END IF

      CALL section_vals_val_get(smeagol_section, "AM.AtomListBS", explicit=is_explicit)
      IF (is_explicit) THEN
         CALL section_vals_val_get(smeagol_section, "AM.AtomListBS", i_vals=iptr)

         ! subsys is not available at this stage, so we cannot check that MINVAL(iptr) <= natoms so far
         IF (SIZE(iptr) == 0 .OR. MINVAL(iptr) <= 0) THEN
            CALL cp_abort(__LOCATION__, &
                          "All atomic indices in AM.AtomListBS should be positive integer numbers.")
         END IF

         n = SIZE(iptr)
         ALLOCATE (ivec(n), indices(n))
         ivec(1:n) = iptr(1:n)
         CALL sort(ivec, n, indices)

         n_unique = 1
         DO i = 2, n
            IF (ivec(i) > ivec(i - 1)) n_unique = n_unique + 1
         END DO

         ALLOCATE (smeagol_control%atomlist_bs(n_unique))
         n_unique = 1
         smeagol_control%atomlist_bs(1) = ivec(1)

         DO i = 2, n
            IF (ivec(i) > ivec(i - 1)) THEN
               n_unique = n_unique + 1
               smeagol_control%atomlist_bs(n_unique) = ivec(i)
            END IF
         END DO
      END IF

      CALL timestop(handle)
   END SUBROUTINE read_smeagol_aux_control
END MODULE smeagol_control_types
